(* csx.sml:
 *  Modified from calc.sml. 
 *  Daniel Phelps
 *)


(* This file provides glue code for building the calculator using the
 * parser and lexer specified in calc.lex and calc.grm.
*)

structure CSX : sig
               val compile : string -> int 
               val usage : unit -> unit
               val main : (string * string list -> int)
                 end = 
struct

open TextIO;
open List;
exception TypeError;
(* 
 * We apply the functors generated from calc.lex and calc.grm to produce
 * the CSXParser structure.
 *)

  structure CSXLrVals =
    CSXLrValsFun(structure Token = LrParser.Token)

  structure CSXLex =
    CSXLexFun(structure Tokens = CSXLrVals.Tokens)

  structure CSXParser =
    Join(structure LrParser = LrParser
	 structure ParserData = CSXLrVals.ParserData
	 structure Lex = CSXLex)

 fun class_name(path) = 
   let val filename = last(String.tokens (fn x => x = #"/") path)
       val no_ext = hd(String.tokens (fn x => x = #".") filename) in
       no_ext
   end

   fun usage() =  (
     print("usage: csx [csx source file]")
   )

   fun exnToString(e) =
      List.foldr (op ^) "" ["[",
                            exnName(e),
                            if exnName(e) = exnMessage(e) then "" else " "^exnMessage(e),
                            "]"]

  fun println(s) = print(s ^ "\n");

  fun file_exists(file) = (
        let val s = openIn(file) in 
          closeIn(s);
	  true
        end
        handle e => (println(exnMessage(e)); false)
  )
  
   fun do_compile(file) = (
     let val (parse_tree, _) = (Parser.parse(file)) 
         val (any_errors, checked_tree) = TypeCheck.check(parse_tree, file) in
       (
       if any_errors then(
         raise TypeError 
       )
       else 
          let val code = CodeGen.gen(checked_tree)
          val s = openOut(class_name(file)^".j") in
                  output(s, code);
                  closeOut(s);
                  0 
         end
         )
     end
   )

 fun compile(file) = (
   if file_exists(file) then
       do_compile(file)
   else (
       usage();
       1
   )
       
 )


   fun main(prog_name, args) = (
     if length(args) = 2 then (
         let val rv = compile(hd(tl(args))) handle e => (print("Compilation failed: "^exnToString(e)^"\n"); 1)in (
            rv
         )
         end
     )
     else (
       usage();
       1
     )
   )
   handle TypeError => (1)
   handle e => (print("CSX compiler bug: " ^ exnToString(e) ^ "\n");
                2)

end (* structure CSX *)
