structure SymbolTable =  
struct 
  open HashTable;
  open DataTypes;

  datatype jvm_address = StaticField
                       | MethodAddress
                       | StackVar of int 

  datatype label_stack = LabelStack of label list

  datatype symb_info = CSXMethod 
                     | CSXField
                     | CSXLocalDecl
                     | CSXFormalArg
                     | CSXLabel
                     | CSXClass of (string)
                     | NoneSymbInfo
                     | Info of {name : string, tipe : csx_type, decl : csx_decl, whatisit : symb_info}
                     | CodegenInfo of {info : symb_info, addr : jvm_address}

  datatype current_method = CurrentMethod of id 
                          | NoCurrentMethod

  datatype scope = Scope of { table : (string, symb_info) hash_table}

  datatype symb_table = SymbTable of (scope list * label_stack * current_method)

  fun replace_current_method(SymbTable(scopes, labels, old_method), new_method : current_method) = 
    SymbTable(scopes, labels, new_method)

  exception DuplicateSymb;
  exception NoAvailableScope;
  exception LabelStackEmpty;
  exception LabelNotFound;
  exception NotFound;
  exception NotTraversingAMethod;
  exception NotAMethod;

  fun add_label(Label(id(name)), SymbTable(scopes, LabelStack(nil), method)) = raise LabelStackEmpty
    | add_label(Label(id(name)), SymbTable(scopes, LabelStack(labels), method)) = 
        SymbTable(scopes, LabelStack(Label(id(name))::labels), method) 
    | add_label(LabelBeginEnd(id(name), begin_label, end_label), SymbTable(scopes, LabelStack(labels), method)) = 
        SymbTable(scopes, LabelStack(LabelBeginEnd(id(name), begin_label, end_label)::labels), method) 
    | add_label(NoneLabel, table) = table

  val hash_fn : string->word = HashString.hashString
  fun cmp_fn (a : string, b : string) = (a = b)
  val initial_size : int = 50  (* I dunno. *)

  fun make_hash_table() = (mkTable (hash_fn, cmp_fn) 
                                             (initial_size, NotFound))
  fun make_scope() = Scope({table = make_hash_table()});

  fun open_scope(SymbTable(nil, labels, method) : symb_table) = 
      SymbTable([make_scope()], labels, method)
    | open_scope(SymbTable(top::scopes, labels, method)) = 
      SymbTable(make_scope()::([top] @ scopes), labels, method)

  fun make_symbol_table() = SymbTable([make_scope()], LabelStack([]), NoCurrentMethod);

  fun close_scope(SymbTable(Scope(top)::rest, labels, method)) = SymbTable(rest, labels, method)
    | close_scope(SymbTable(nil, labels, method)) = raise NoAvailableScope

  fun insert_symb(_,_, SymbTable(nil, labels, method)) = raise NoAvailableScope
    | insert_symb(name : string, 
                  info : symb_info, 
                  SymbTable(Scope(top)::rest, labels, method) : symb_table) =
        let val the_table = (#table(top))
        in 
          if inDomain (the_table) name then
            raise DuplicateSymb
          else
            let val _ = insert the_table (name, info) in (
              SymbTable(Scope({table = the_table})::rest, labels, method)
              )
            end
        end
              
  fun get_symb(name : string, SymbTable(nil, _, _)) = (raise NotFound)
        
    | get_symb(name : string, SymbTable(Scope(top)::rest, labels, method)) = (
        (*print("lookin for " ^name);*)
        let val the_table = (#table(top))
        in 
          case find the_table (name) of SOME info => info
                                      | NONE => (get_symb(name, SymbTable(rest, labels, method)))
        end
        )

  fun args_of(info : symb_info) =  
    case info of Info({name : string, tipe : csx_type, decl : csx_decl, whatisit = CSXMethod}) => method_args(decl)
                      | _ => (raise NotAMethod)

  fun return_type_of(method : symb_info) =  (
   
    case method of Info({name : string, tipe : csx_type, decl : csx_decl, 
      whatisit = CSXMethod}) 
                       => method_return_type(decl)
               | _ => (raise NotAMethod)
         )

  fun is_const(var : symb_info) = 
    case var of Info({name : string, tipe = Const(t), decl : csx_decl, 
                       whatisit = CSXField}) 
                       => true
               | _ => false

  fun get_return_type_of(method : string, table : symb_table) = 
      let val s = get_symb(method, table) in (
        ( 
          return_type_of(s))
        )
      end

  fun is_method(name : string, table : symb_table) =  (
    let val return_type = get_return_type_of(name, table) in
        true
    end
    handle NotAMethod => (false)
         | NotFound => (false)
         | _ => (print("this is messed up\n"); false)
 )

  fun get_method_args(name, table) = args_of(get_symb(name, table))
    
  fun get_current_return_type(SymbTable(scopes, labels, NoCurrentMethod)) =
                raise NotTraversingAMethod
    | get_current_return_type(SymbTable(scopes, labels, CurrentMethod(id(s)))) =
    (
       get_return_type_of(s, SymbTable(scopes, labels, CurrentMethod(id(s))))
     )

  fun get_current_method(SymbTable(scopes, labels, NoCurrentMethod)) = 
        raise NotTraversingAMethod
    | get_current_method(SymbTable(scopes, labels, CurrentMethod(id(s)))) = 
        get_symb(s, SymbTable(scopes, labels, CurrentMethod(id(s))))

  fun symb_exists(name, table) = 
    let val info = get_symb(name, table) in
      true
    end
    handle NotFound => false

  fun label_exists(name, table) = 
    let val info = get_symb(name, table) in
      case info of CSXLabel => true
                 | _        => false 
    end
    handle NotFound => false

  fun get_label_begin_end(name, SymbTable(_, LabelStack(nil), _)) = raise LabelNotFound 
    | get_label_begin_end(name, SymbTable(scopes, LabelStack(label::labels), method)) = 
    let val search_rest = (fn () => get_label_begin_end(name, SymbTable(scopes, LabelStack(labels), method)))
    in (
      case label of LabelBeginEnd(id(s), begin_label, end_label) => if s = name then 
                                                                   (begin_label, end_label) 
                                                                else 
                                                                   search_rest()
                  | _ => search_rest()
    ) end

  fun type_of(name, table) = 
          let val info = get_symb(name, table) in
            case info of Info({name : string, tipe : csx_type, decl : csx_decl, whatisit : symb_info}) => tipe
                       | _ => NoneType 
          end

  fun decl_type(FormalArg(tipe, name, _, _), table) = (
        tipe
  )

  fun open_scope_and_then(table, action) = 
    let val augmented_table = open_scope(table) in
      action(augmented_table)
    end

  fun debug(SymbTable(nil, _, _)) = nil
    | debug(SymbTable(Scope(top)::rest, labels, method)) = 
        listItemsi(#table(top))::debug(SymbTable(rest, labels, method))
end;
